/**
 * @author santa
 * @created Nov 14, 2008
 */
package search.maze;

import java.awt.Point;
import java.awt.Dimension;
import java.util.ArrayList;


public class Maze {

    public static final int PATH = 2;
    public static final int VISITED = 1;
    public static final int FREE = 0;
    public static final int START = -1;
    public static final int GOAL = -2;
    public static final int OBSTACLE = -3;

    private Point start;
    private Point goal;
    private Dimension dimension;
    private int[][] maze;

    private ArrayList<MazeListener> mazeListeners;

    public Maze(int width, int height) {
        this(new Dimension(width, height), new Point(0, 0), new Point(
                width - 1, height - 1));

        int sx = (int) (Math.random() * width);
        int sy = (int) (Math.random() * height);
        int gx, gy;
        do {
            gx = (int) (Math.random() * width);
            gy = (int) (Math.random() * height);
        } while (gx == sx && gy == sy);

        start = new Point(sx, sy);
        goal = new Point(gx, gy);
        resetMaze();
    }

    public Maze(Dimension d, Point s, Point g) {
        dimension = d;
        start = s;
        goal = g;
        maze = new int[d.width][d.height];

        // create a maze with randomized obstacles
        for (int i = 0; i < (d.width * d.height / 3); i++) {
            int x = (int) (Math.random() * d.width);
            int y = (int) (Math.random() * d.height);
            setValue(x, y, OBSTACLE);
        }

        setValue(s.x, s.y, START);
        setValue(g.x, g.y, GOAL);
    }

    public Maze(Dimension d, Point s, Point g, int[][] m) {
        dimension = d;
        start = s;
        goal = g;
        maze = m;

        setValue(s.x, s.y, START);
        setValue(g.x, g.y, GOAL);
    }

    public void resetMaze() {
        for (int y = 0; y < dimension.height; y++) {
            for (int x = 0; x < dimension.width; x++) {
                int value = getValue(x, y);
                switch (value) {
                case OBSTACLE:
                    setValue(x, y, value);
                    break;
                default:
                    setValue(x, y, FREE);
                }
            }
        }
        setValue(start.x, start.y, START);
        setValue(goal.x, goal.y, GOAL);
    }

    public void addMazeListener(MazeListener listener) {
        if (mazeListeners == null)
            mazeListeners = new ArrayList<MazeListener>();

        mazeListeners.add(listener);
    }

    public void removeMazeListener(MazeListener listener) {
        if (mazeListeners == null)
            return;

        mazeListeners.remove(listener);
    }

    public void notifyListeners() {
        if (mazeListeners == null)
            return;

        for (MazeListener listener : mazeListeners)
            listener.mazeChanged(this);
    }

    public int getWidth() {
        return dimension.width;
    }

    public int getHeight() {
        return dimension.height;
    }

    public Point getStart() {
        return start;
    }

    public Point getGoal() {
        return goal;
    }

    public synchronized int getValue(int x, int y) {
        int value;

        try {
            value = maze[y][x];
        } catch (ArrayIndexOutOfBoundsException e) {
            value = OBSTACLE;
        }

        return value;
    }

    public synchronized void setValue(int x, int y, int value) {
        maze[y][x] = value;
        notifyListeners();
    }

    public Point[] getPossibleMoves(Point location) {
        ArrayList<Point> moves = new ArrayList<Point>();
        Point[] neighbors = new Point[4];

        neighbors[0] = new Point(location.x - 1, location.y);
        neighbors[1] = new Point(location.x + 1, location.y);
        neighbors[2] = new Point(location.x, location.y - 1);
        neighbors[3] = new Point(location.x, location.y + 1);

        for (int i = 0; i < neighbors.length; i++) {
            int value = getValue(neighbors[i].x, neighbors[i].y);

            switch (value) {
            case Maze.FREE:
            case Maze.GOAL:
                moves.add(neighbors[i]);
                break;
            default:
            }
        }

        Point[] ret = new Point[moves.size()];
        return moves.toArray(ret);
    }

    // for debugging
    public String toString() {
        StringBuffer sb = new StringBuffer();
        for (int y = 0; y < getHeight(); y++) {
            for (int x = 0; x < getWidth(); x++) {
                int value = getValue(x, y);
                switch (value) {
                case OBSTACLE:
                    sb.append("B ");
                    break;
                case START:
                    sb.append("S ");
                    break;
                case GOAL:
                    sb.append("G ");
                    break;
                case FREE:
                    sb.append(". ");
                    break;
                case PATH:
                    sb.append("P ");
                    break;
                default:
                    sb.append("V ");
                    break;
                }
            }
            sb.append("\n");
        }

        return sb.toString();
    }

    // for unit test
    public static void main(String[] args) {
        System.out.println(new Maze(10, 10));
        System.exit(0);
    }

}
