/**
 * @author santa
 * @created Nov 14, 2008
 */
package search.maze;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import org.jdesktop.swingworker.SwingWorker;


@SuppressWarnings("serial")
public class MazeSearch extends JPanel implements ActionListener, MazeListener {

    Maze maze;
    SearchEngine searchEngine;

    MazeGUI mazeGUI;
    JComboBox algorithms;
    JButton search, reset, random;

    public MazeSearch() {
        super();
        maze = new Maze(10, 10);
        maze.addMazeListener(this);
        try {
            SwingUtilities.invokeAndWait(new Runnable() {

                public void run() {
                    initGUI();
                }
            });
        } catch (Exception e) {
            System.err.println("Error initializing GUI.");
            System.exit(-1);
        }
    }

    private void initGUI() {
        setPreferredSize(new Dimension(420, 450));
        mazeGUI = new MazeGUI();
        mazeGUI.setPreferredSize(new Dimension(400, 400));

        String[] algorithmNames = { "DFS", "BFS", "A*" };
        algorithms = new JComboBox(algorithmNames);
        algorithms.setSelectedIndex(0);
        search = new JButton("Search");
        reset = new JButton("Reset");
        random = new JButton("Random");
        JPanel optionPane = new JPanel();

        algorithms.addActionListener(this);
        search.addActionListener(this);
        reset.addActionListener(this);
        random.addActionListener(this);

        optionPane.setLayout(new FlowLayout());
        optionPane.add(algorithms);
        optionPane.add(search);
        optionPane.add(reset);
        optionPane.add(random);

        setLayout(new BorderLayout());
        add(mazeGUI, BorderLayout.CENTER);
        add(optionPane, BorderLayout.SOUTH);
    }

    private void resetMaze() {
        searchEngine = null;
        maze.resetMaze();
    }

    private void randomMaze() {
        searchEngine = null;
        maze = new Maze(10, 10);
        maze.addMazeListener(this);
        maze.notifyListeners();
    }

    public void actionPerformed(ActionEvent evt) {
        Object source = evt.getSource();

        if (source == algorithms) {
            // pass?
        } else if (source == search) {
            search.setEnabled(false);
            random.setEnabled(false);
            reset.setEnabled(false);

            String algorithmName = (String) algorithms.getSelectedItem();
            startSearch(algorithmName);
        } else if (source == reset) {
            resetMaze();
        } else if (source == random) {
            randomMaze();
        }
    }

    private void startSearch(String algorithm) {
        resetMaze();

        if (algorithm.equals("DFS"))
            searchEngine = new DFSEngine(maze);
        else if (algorithm.equals("BFS"))
            searchEngine = new BFSEngine(maze);
        else if (algorithm.equals("A*"))
            searchEngine = new AStarEngine(maze);

        SwingWorker<Void, Void> worker = new SwingWorker<Void, Void>() {

            @Override
            public Void doInBackground() throws Exception {
                try {
                    searchEngine.search();
                } catch (PathNotFoundException ex) {
                    System.err.println("Path not found.");
                    resetMaze();
                }

                return null;
            }

            @Override
            public void done() {
                mazeChanged(maze);
                if (searchEngine != null)
                    System.out.println("Path found.");
            }

        };

        worker.execute();
    }

    public void mazeChanged(Maze src) {
        mazeGUI.repaint();

        if (searchEngine == null || !searchEngine.searching) {
            search.setEnabled(true);
            random.setEnabled(true);
            reset.setEnabled(true);
        } else {
            search.setEnabled(false);
            random.setEnabled(false);
            reset.setEnabled(false);
        }
    }

    private class MazeGUI extends JPanel {

        Image offscreen;

        public void paint(Graphics g) {
            Dimension dim = getSize();
            checkOffscreenImage();
            Graphics gr = offscreen.getGraphics();
            gr.setColor(getBackground());
            gr.fillRect(0, 0, dim.width, dim.height);

            render(offscreen.getGraphics());
            g.drawImage(offscreen, 0, 0, null);
        }

        private void checkOffscreenImage() {
            Dimension dim = getSize();
            if (offscreen == null || offscreen.getWidth(null) != dim.width
                    || offscreen.getHeight(null) != dim.height) {
                offscreen = createImage(dim.width, dim.height);
            }
        }

        public void render(Graphics g) {
            Dimension dim = getSize();

            g.setColor(Color.WHITE);
            g.fillRect(0, 0, dim.width, dim.height);

            int unitWidth = (dim.width - maze.getWidth()) / maze.getWidth();
            int unitHeight = (dim.height - maze.getHeight()) / maze.getHeight();
            int widthOffset = maze.getWidth() / 2;
            int heightOffset = maze.getHeight() / 2;
            for (int x = 0; x < maze.getWidth(); x++) {
                for (int y = 0; y < maze.getHeight(); y++) {
                    switch (maze.getValue(x, y)) {
                    case Maze.OBSTACLE:
                        g.setColor(Color.DARK_GRAY);
                        break;
                    case Maze.PATH:
                        g.setColor(Color.CYAN);
                        break;
                    case Maze.VISITED:
                        g.setColor(Color.LIGHT_GRAY);
                        break;
                    default:
                        g.setColor(Color.WHITE);
                        break;
                    }
                    g.fillRect(x * unitWidth + widthOffset, y * unitHeight
                            + heightOffset, unitWidth, unitHeight);
                    g.setColor(Color.BLACK);
                    g.drawRect(x * unitWidth + widthOffset, y * unitHeight
                            + heightOffset, unitWidth, unitHeight);
                }
            }

            // always draw START and GOAL cells
            g.setFont(g.getFont().deriveFont(15.0f));
            g.setColor(Color.RED);
            FontMetrics fmetric = g.getFontMetrics();
            int fontWidth;
            int fontAscent = fmetric.getAscent();

            fontWidth = fmetric.charWidth('S');
            g.drawString("S", maze.getStart().x * unitWidth
                    + (unitWidth - fontWidth) / 2 + widthOffset,
                    maze.getStart().y * unitHeight
                    + (unitHeight + fontAscent) / 2 + heightOffset);

            fontWidth = fmetric.charWidth('G');
            g.drawString("G", maze.getGoal().x * unitWidth
                    + (unitWidth - fontWidth) / 2 + widthOffset,
                    maze.getGoal().y * unitHeight 
                    + (unitHeight + fontAscent) / 2 + heightOffset);

            g.dispose();
        }

    }

}
