/**
 * @author santa
 * @created Nov 14, 2008
 */
package search.maze;

import java.awt.Point;
import java.util.ArrayList;
import java.util.Collections;


/**
 * An abstract superclass to search engine objects employing some informed
 * search algorithm and path finding. Subclasses need to implement the abstract
 * method <code>doSearch()</code> and must return either <code>null</code> if it
 * cannot find a path, or a <code>Node</code> object that represents the maze's
 * goal.
 * 
 * @author santa
 */
public abstract class SearchEngine {

    protected Maze maze;
    protected ArrayList<Point> path;
    protected boolean searching;
    protected int delay;

    public SearchEngine(Maze m) {
        maze = m;
        path = null;
        searching = false;
        delay = 250;
    }

    public void setDelay(int delay) {
        this.delay = delay;
    }

    public Maze getMaze() {
        return maze;
    }

    public synchronized Point[] getPath() {
        if (searching || path == null)
            return null;

        Point[] rpath = new Point[path.size()];
        for (int i = 0; i < path.size(); i++)
            rpath[i] = path.get(i);

        return rpath;
    }

    public void search() throws PathNotFoundException {
        path = null;
        searching = true;
        Node result = doSearch();
        searching = false;
        if (result == null)
            throw new PathNotFoundException();
        else
            constructPath(result);
    }

    private void constructPath(Node node) {
        ArrayList<Point> path = new ArrayList<Point>();

        while (node != null) {
            path.add(node.ref);
            maze.setValue(node.ref.x, node.ref.y, Maze.PATH);
            try {
                Thread.sleep(delay / 10);
            } catch (InterruptedException e) {
            }
            node = node.predecessor;
        }

        Collections.reverse(path);
        this.path = path;
    }

    protected abstract Node doSearch();

    protected class Node implements Comparable<Node> {

        protected Point ref;
        protected Node predecessor;

        protected Node(Point p, Node pre) {
            ref = p;
            predecessor = pre;
        }

        private int distanceTravelled() {
            int distance = 0;
            Node ptr = this;
            while (ptr.predecessor != null) {
                distance++;
                ptr = ptr.predecessor;
            }
            return distance;
        }

        /**
         * Returns the heuristic value of this node, which by default is the sum
         * of the distance travelled so far by this node, and its Manhattan
         * distance to the goal.
         * 
         * @return the heuristic value of this node
         */
        protected int getHeuristic() {
            return distanceTravelled() + Math.abs(ref.x - maze.getGoal().x)
                    + Math.abs(ref.y - maze.getGoal().y);
        }

        public int compareTo(Node n) {
            return this.getHeuristic() - n.getHeuristic();
        }

    }

}
